###########################
### query_processing.py ###
###########################
# Author: Samuel Aroney

import polars as pl
import os

OUTPUT_COLUMNS={
    "gene": str,
    "sample": str,
    "sequence": str,
    "num_hits": int,
    "coverage": int,
    "taxonomy": str,
    "found_in": str,
    }

def processing(
    query_read,
    pipe_read,
    SEQUENCE_IDENTITY=0.86,
    WINDOW_SIZE=60,
    TAXA_OF_INTEREST=None):

    if len(query_read) == 0:
        empty_output = pl.DataFrame(schema=OUTPUT_COLUMNS)
        return empty_output, empty_output

    # Filter TAXA_OF_INTEREST
    if TAXA_OF_INTEREST:
        pipe_read = pipe_read.filter(
            pl.col("taxonomy").str.contains(TAXA_OF_INTEREST, literal=True)
        )

    appraised = query_read.rename(
        # Rename to match appraise output
        # Query output: query_name, query_sequence, divergence, num_hits, coverage, sample, marker, hit_sequence, taxonomy
        # Appraise output: gene, sample, sequence, num_hits, coverage, taxonomy, found_in
        {"marker": "gene", "query_name":"sample", "query_sequence": "sequence", "sample": "found_in"}
    ).drop([
        # Query taxonomy, num_hits and coverage are from database (e.g. the genomes)
        "taxonomy", "num_hits", "coverage"
    ]).join(
        pipe_read, on=["gene", "sample", "sequence"], how="inner"
    ).groupby(
        ["gene", "sample", "sequence", "num_hits", "coverage", "taxonomy", "divergence"]
    ).agg(
        pl.col("found_in").sort().str.concat(",")
    ).with_columns(
        pl.col("divergence").alias("binned") <= ((1 - SEQUENCE_IDENTITY) * WINDOW_SIZE)
    )

    # Split dataframe into binned/unbinned
    binned = appraised.filter(pl.col("binned")).drop(["divergence", "binned"])
    unbinned = pipe_read.join(
        appraised, on=["gene", "sample", "sequence", "num_hits", "coverage", "taxonomy"], how="left"
    ).filter(~pl.col("binned").fill_null(False)
    ).drop(["divergence", "binned"]
    ).with_columns(
        pl.lit(None).cast(str).alias("found_in")
    )

    # Filter out EIF
    binned = binned.filter(pl.col("gene") != "S3.18.EIF_2_alpha")
    unbinned = unbinned.filter(pl.col("gene") != "S3.18.EIF_2_alpha")

    return binned, unbinned

def pipeline(
    query_reads,
    pipe_reads,
    SEQUENCE_IDENTITY=0.86,
    WINDOW_SIZE=60,
    TAXA_OF_INTEREST=None):

    print(f"Polars using {str(pl.threadpool_size())} threads")

    for query, pipe in zip(query_reads, pipe_reads):
        binned, unbinned = processing(
            pl.read_csv(query, separator="\t"),
            pl.read_csv(pipe, separator="\t"),
            SEQUENCE_IDENTITY,
            WINDOW_SIZE,
            TAXA_OF_INTEREST)
        yield binned, unbinned

if __name__ == "__main__":
    os.environ["POLARS_MAX_THREADS"] = str(snakemake.threads)
    import polars as pl

    SEQUENCE_IDENTITY = snakemake.params.sequence_identity
    WINDOW_SIZE = snakemake.params.window_size
    TAXA_OF_INTEREST = snakemake.params.taxa_of_interest
    query_reads = snakemake.input.query_reads
    pipe_reads = snakemake.input.pipe_reads
    binned_path = snakemake.output.binned
    unbinned_path = snakemake.output.unbinned

    outputs = pipeline(
        query_reads,
        pipe_reads,
        SEQUENCE_IDENTITY=SEQUENCE_IDENTITY,
        WINDOW_SIZE=WINDOW_SIZE,
        TAXA_OF_INTEREST=TAXA_OF_INTEREST
        )

    with open(binned_path, "ab") as binned_file, open(unbinned_path, "ab") as unbinned_file:
        first = True
        for binned, unbinned in outputs:
            binned.write_csv(binned_file, separator="\t", has_header=first)
            unbinned.write_csv(unbinned_file, separator="\t", has_header=first)
            first = False



#!/usr/bin/env python3

import unittest
import os
os.environ["POLARS_MAX_THREADS"] = "1"
import polars as pl
from polars.testing import assert_frame_equal
from ibis.workflow.scripts.evaluate import evaluate

SINGLEM_COLUMNS=["gene", "sample", "sequence", "num_hits", "coverage", "taxonomy"]
TARGET_COLUMNS=SINGLEM_COLUMNS+["target"]
APPRAISE_COLUMNS=SINGLEM_COLUMNS+["found_in"]
CLUSTER_COLUMNS=["samples", "length", "total_targets", "total_size", "recover_samples", "coassembly"]
EDGE_COLUMNS=["style", "cluster_size", "samples", "target_ids"]

OUTPUT_COLUMNS={
    "coassembly": str,
    "gene": str,
    "sequence": str,
    "genome": str,
    "target": str,
    "found_in": str,
    "source_samples": str,
    "taxonomy": str,
    }
SUMMARY_COLUMNS = {
    "coassembly": str,
    "statistic": str,
    "within": str,
    "match": int,
    "nonmatch": int,
    "total": int,
    "match_percent": float,
    }

class Tests(unittest.TestCase):
    def assertDataFrameEqual(self, a, b):
        assert_frame_equal(a, b, check_dtype=False, check_row_order=False)

    def test_evaluate_script(self):
        targets = pl.DataFrame([
            ["S3.1", "sample_1.1", "AAA", 5, 10.0, "Root; old", 10],
            ["S3.1", "sample_2.1", "AAA", 5, 10.0, "Root; old", 10],
            ["S3.1", "sample_1.1", "CCC", 1, 2.0, "Root; old", 11],
            ["S3.1", "sample_2.1", "CCC", 1, 2.0, "Root; old", 11],
            ["S3.1", "sample_1.1", "DDD", 1, 2.0, "Root; old", 12],
        ], schema=TARGET_COLUMNS)
        binned = pl.DataFrame([
            ["S3.1", "sample_1.1", "BBB", 5, 10.0, "Root; old", "oldgenome_1"],
        ], schema=APPRAISE_COLUMNS)
        clusters = pl.DataFrame([
            ["sample_1,sample_2", 2, 1, 100, "sample_1,sample_2,sample_3", "coassembly_0"],
        ], schema=CLUSTER_COLUMNS)
        edges = pl.DataFrame([
            ["match", 2, "sample_1.1,sample_2.1", "10"],
        ], schema=EDGE_COLUMNS)
        recovered = pl.DataFrame([
            ["S3.1", "coassembly_0-genome_1_transcripts", "AAA", 1, 2.0, "Root"],
            ["S3.1", "coassembly_0-genome_1_transcripts", "AAB", 1, 2.0, "Root"],
            ["S3.1", "coassembly_0-genome_1_transcripts", "BBB", 1, 2.0, "Root"],
            ["S3.1", "coassembly_0-genome_1_transcripts", "CCC", 1, 2.0, "Root"],
            ["S3.1", "coassembly_0-genome_1_transcripts", "DDD", 1, 2.0, "Root"],
        ], schema=SINGLEM_COLUMNS)
        recovered_bins = {
            "coassembly_0-genome_0": "genome_0.fna",
            "coassembly_0-genome_1": "genome_1.fna",
            }

        expected_matches = pl.DataFrame([
            ["coassembly_0", "S3.1", "AAA", "genome_1_transcripts", "10", None, "sample_1,sample_2", "Root"],
            ["coassembly_0", "S3.1", "BBB", "genome_1_transcripts", None, "oldgenome_1", "sample_1", "Root"],
            ["coassembly_0", "S3.1", "CCC", "genome_1_transcripts", None, None, "sample_1,sample_2", "Root"],
            ["coassembly_0", "S3.1", "DDD", "genome_1_transcripts", None, None, "sample_1", "Root"],
        ], schema=OUTPUT_COLUMNS)
        expected_unmatched = pl.DataFrame([
            ["coassembly_0", "S3.1", "AAB", "genome_1_transcripts", None, None, None, "Root"],
        ], schema=OUTPUT_COLUMNS)
        expected_summary = pl.DataFrame([
            ["coassembly_0", "sequences", "targets", 1, 0, 1, 100],
            ["coassembly_0", "taxonomy", "targets", 1, 0, 1, 100],
            ["coassembly_0", "nontarget_bin_sequences", "recovery", 1, 4, 5, 20.00],
            ["coassembly_0", "nontarget_unbin_sequences", "recovery", 2, 3, 5, 40.00],
            ["coassembly_0", "novel_sequences", "recovery", 1, 4, 5, 20.00],
            ["coassembly_0", "bins", "recovery", 1, 1, 2, 50],
        ], schema=SUMMARY_COLUMNS)

        observed_matches, observed_unmatched, observed_summary = evaluate(targets, binned, clusters, edges, recovered, recovered_bins)
        self.assertDataFrameEqual(expected_matches, observed_matches)
        self.assertDataFrameEqual(expected_unmatched, observed_unmatched)
        self.assertDataFrameEqual(expected_summary, observed_summary)

    def test_evaluate_script_all_targets(self):
        targets = pl.DataFrame([
            ["S3.1", "sample_1.1", "AAA", 5, 10.0, "Root; old", 10],
            ["S3.1", "sample_2.1", "AAA", 5, 10.0, "Root; old", 10],
        ], schema=TARGET_COLUMNS)
        binned = pl.DataFrame([
            ["S3.1", "sample_1.1", "BBB", 5, 10.0, "Root; old", "oldgenome_1"],
        ], schema=APPRAISE_COLUMNS)
        clusters = pl.DataFrame([
            ["sample_1,sample_2", 2, 1, 100, "sample_1,sample_2,sample_3", "coassembly_0"],
        ], schema=CLUSTER_COLUMNS)
        edges = pl.DataFrame([
            ["match", 2, "sample_1.1,sample_2.1", "10"],
        ], schema=EDGE_COLUMNS)
        recovered = pl.DataFrame([
            ["S3.1", "coassembly_0-genome_1_transcripts", "AAA", 1, 2.0, "Root"],
        ], schema=SINGLEM_COLUMNS)
        recovered_bins = {
            "coassembly_0-genome_0": "genome_0.fna",
            "coassembly_0-genome_1": "genome_1.fna",
            }

        expected_matches = pl.DataFrame([
            ["coassembly_0", "S3.1", "AAA", "genome_1_transcripts", "10", None, "sample_1,sample_2", "Root"],
        ], schema=OUTPUT_COLUMNS)
        expected_unmatched = pl.DataFrame([
        ], schema=OUTPUT_COLUMNS)
        expected_summary = pl.DataFrame([
            ["coassembly_0", "sequences", "targets", 1, 0, 1, 100],
            ["coassembly_0", "taxonomy", "targets", 1, 0, 1, 100],
            ["coassembly_0", "nontarget_bin_sequences", "recovery", 0, 1, 1, 0],
            ["coassembly_0", "nontarget_unbin_sequences", "recovery", 0, 1, 1, 0],
            ["coassembly_0", "novel_sequences", "recovery", 0, 1, 1, 0],
            ["coassembly_0", "bins", "recovery", 1, 1, 2, 50],
        ], schema=SUMMARY_COLUMNS)

        observed_matches, observed_unmatched, observed_summary = evaluate(targets, binned, clusters, edges, recovered, recovered_bins)
        self.assertDataFrameEqual(expected_matches, observed_matches)
        self.assertDataFrameEqual(expected_unmatched, observed_unmatched)
        self.assertDataFrameEqual(expected_summary, observed_summary)

    def test_evaluate_script_none_targets(self):
        targets = pl.DataFrame([
            ["S3.1", "sample_1.1", "AAA", 5, 10.0, "Root; old", 10],
            ["S3.1", "sample_2.1", "AAA", 5, 10.0, "Root; old", 10],
        ], schema=TARGET_COLUMNS)
        binned = pl.DataFrame([
            ["S3.1", "sample_1.1", "BBB", 5, 10.0, "Root; old", "oldgenome_1"],
        ], schema=APPRAISE_COLUMNS)
        clusters = pl.DataFrame([
            ["sample_1,sample_2", 2, 1, 100, "sample_1,sample_2,sample_3", "coassembly_0"],
        ], schema=CLUSTER_COLUMNS)
        edges = pl.DataFrame([
            ["match", 2, "sample_1.1,sample_2.1", "10"],
        ], schema=EDGE_COLUMNS)
        recovered = pl.DataFrame([
            ["S3.1", "coassembly_0-genome_1_transcripts", "AAB", 1, 2.0, "Root"],
        ], schema=SINGLEM_COLUMNS)
        recovered_bins = {
            "coassembly_0-genome_0": "genome_0.fna",
            "coassembly_0-genome_1": "genome_1.fna",
            }

        expected_matches = pl.DataFrame([
            ["coassembly_0", "S3.1", "AAA", None, "10", None, "sample_1,sample_2", "Root; old"],
        ], schema=OUTPUT_COLUMNS)
        expected_unmatched = pl.DataFrame([
            ["coassembly_0", "S3.1", "AAB", "genome_1_transcripts", None, None, None, "Root"],
        ], schema=OUTPUT_COLUMNS)
        expected_summary = pl.DataFrame([
            ["coassembly_0", "sequences", "targets", 0, 1, 1, 0],
            ["coassembly_0", "taxonomy", "targets", 0, 1, 1, 0],
            ["coassembly_0", "nontarget_bin_sequences", "recovery", 0, 1, 1, 0],
            ["coassembly_0", "nontarget_unbin_sequences", "recovery", 0, 1, 1, 0],
            ["coassembly_0", "novel_sequences", "recovery", 1, 0, 1, 100],
            ["coassembly_0", "bins", "recovery", 0, 2, 2, 0],
        ], schema=SUMMARY_COLUMNS)

        observed_matches, observed_unmatched, observed_summary = evaluate(targets, binned, clusters, edges, recovered, recovered_bins)
        self.assertDataFrameEqual(expected_matches, observed_matches)
        self.assertDataFrameEqual(expected_unmatched, observed_unmatched)
        self.assertDataFrameEqual(expected_summary, observed_summary)

    def test_evaluate_script_empty_recovery(self):
        targets = pl.DataFrame([
            ["S3.1", "sample_1.1", "AAA", 5, 10.0, "Root; old", 10],
            ["S3.1", "sample_2.1", "AAA", 5, 10.0, "Root; old", 10],
            ["S3.1", "sample_1.1", "CCC", 1, 2.0, "Root; old", 11],
            ["S3.1", "sample_2.1", "CCC", 1, 2.0, "Root; old", 11],
        ], schema=TARGET_COLUMNS)
        binned = pl.DataFrame([
            ["S3.1", "sample_1.1", "BBB", 5, 10.0, "Root; old", "oldgenome_1"],
        ], schema=APPRAISE_COLUMNS)
        clusters = pl.DataFrame([
            ["sample_1,sample_2", 2, 1, 100, "sample_1,sample_2,sample_3", "coassembly_0"],
        ], schema=CLUSTER_COLUMNS)
        edges = pl.DataFrame([
            ["match", 2, "sample_1.1,sample_2.1", "10"],
        ], schema=EDGE_COLUMNS)
        recovered = pl.DataFrame([
        ], schema=SINGLEM_COLUMNS)
        recovered_bins = {
            }

        expected_matches = pl.DataFrame([
        ], schema=OUTPUT_COLUMNS)
        expected_unmatched = pl.DataFrame([
        ], schema=OUTPUT_COLUMNS)
        expected_summary = pl.DataFrame([
        ], schema=SUMMARY_COLUMNS)

        observed_matches, observed_unmatched, observed_summary = evaluate(targets, binned, clusters, edges, recovered, recovered_bins)
        self.assertDataFrameEqual(expected_matches, observed_matches)
        self.assertDataFrameEqual(expected_unmatched, observed_unmatched)
        self.assertDataFrameEqual(expected_summary, observed_summary)

    def test_evaluate_script_nontarget_wrong_sample(self):
        targets = pl.DataFrame([
            ["S3.1", "sample_1.1", "AAA", 5, 10.0, "Root; old", 10],
            ["S3.1", "sample_2.1", "AAA", 5, 10.0, "Root; old", 10],
        ], schema=TARGET_COLUMNS)
        binned = pl.DataFrame([
            ["S3.1", "sample_3.1", "BBB", 5, 10.0, "Root; old", "oldgenome_1"],
        ], schema=APPRAISE_COLUMNS)
        clusters = pl.DataFrame([
            ["sample_1,sample_2", 2, 1, 100, "sample_1,sample_2,sample_3", "coassembly_0"],
        ], schema=CLUSTER_COLUMNS)
        edges = pl.DataFrame([
            ["match", 2, "sample_1.1,sample_2.1", "10"],
        ], schema=EDGE_COLUMNS)
        recovered = pl.DataFrame([
            ["S3.1", "coassembly_0-genome_1_transcripts", "AAA", 1, 2.0, "Root"],
            ["S3.1", "coassembly_0-genome_1_transcripts", "AAB", 1, 2.0, "Root"],
            ["S3.1", "coassembly_0-genome_1_transcripts", "BBB", 1, 2.0, "Root"],
        ], schema=SINGLEM_COLUMNS)
        recovered_bins = {
            "coassembly_0-genome_0": "genome_0.fna",
            "coassembly_0-genome_1": "genome_1.fna",
            }

        expected_matches = pl.DataFrame([
            ["coassembly_0", "S3.1", "AAA", "genome_1_transcripts", "10", None, "sample_1,sample_2", "Root"],
        ], schema=OUTPUT_COLUMNS)
        expected_unmatched = pl.DataFrame([
            ["coassembly_0", "S3.1", "AAB", "genome_1_transcripts", None, None, None, "Root"],
            ["coassembly_0", "S3.1", "BBB", "genome_1_transcripts", None, None, None, "Root"],
        ], schema=OUTPUT_COLUMNS)
        expected_summary = pl.DataFrame([
            ["coassembly_0", "sequences", "targets", 1, 0, 1, 100],
            ["coassembly_0", "taxonomy", "targets", 1, 0, 1, 100],
            ["coassembly_0", "nontarget_bin_sequences", "recovery", 0, 3, 3, 0],
            ["coassembly_0", "nontarget_unbin_sequences", "recovery", 0, 3, 3, 0],
            ["coassembly_0", "novel_sequences", "recovery", 2, 1, 3, 66.67],
            ["coassembly_0", "bins", "recovery", 1, 1, 2, 50],
        ], schema=SUMMARY_COLUMNS)

        observed_matches, observed_unmatched, observed_summary = evaluate(targets, binned, clusters, edges, recovered, recovered_bins)
        self.assertDataFrameEqual(expected_matches, observed_matches)
        self.assertDataFrameEqual(expected_unmatched, observed_unmatched)
        self.assertDataFrameEqual(expected_summary, observed_summary)

    def test_evaluate_script_samples_with_same_sequence(self):
        targets = pl.DataFrame([
            ["S3.1", "sample_1.1", "AAA", 5, 10.0, "Root; old", 10],
            ["S3.1", "sample_2.1", "AAA", 5, 10.0, "Root; old", 10],
        ], schema=TARGET_COLUMNS)
        binned = pl.DataFrame([
            ["S3.1", "sample_1.1", "BBB", 5, 10.0, "Root; old", "oldgenome_1"],
            ["S3.1", "sample_2.1", "BBB", 5, 10.0, "Root; old", "oldgenome_1"],
        ], schema=APPRAISE_COLUMNS)
        clusters = pl.DataFrame([
            ["sample_1,sample_2", 2, 1, 100, "sample_1,sample_2,sample_3", "coassembly_0"],
        ], schema=CLUSTER_COLUMNS)
        edges = pl.DataFrame([
            ["match", 2, "sample_1.1,sample_2.1", "10"],
        ], schema=EDGE_COLUMNS)
        recovered = pl.DataFrame([
            ["S3.1", "coassembly_0-genome_1_transcripts", "AAA", 1, 2.0, "Root"],
            ["S3.1", "coassembly_0-genome_1_transcripts", "AAB", 1, 2.0, "Root"],
            ["S3.1", "coassembly_0-genome_1_transcripts", "BBB", 1, 2.0, "Root"],
        ], schema=SINGLEM_COLUMNS)
        recovered_bins = {
            "coassembly_0-genome_0": "genome_0.fna",
            "coassembly_0-genome_1": "genome_1.fna",
            }

        expected_matches = pl.DataFrame([
            ["coassembly_0", "S3.1", "AAA", "genome_1_transcripts", "10", None, "sample_1,sample_2", "Root"],
            ["coassembly_0", "S3.1", "BBB", "genome_1_transcripts", None, "oldgenome_1", "sample_1,sample_2", "Root"],
        ], schema=OUTPUT_COLUMNS)
        expected_unmatched = pl.DataFrame([
            ["coassembly_0", "S3.1", "AAB", "genome_1_transcripts", None, None, None, "Root"],
        ], schema=OUTPUT_COLUMNS)
        expected_summary = pl.DataFrame([
            ["coassembly_0", "sequences", "targets", 1, 0, 1, 100],
            ["coassembly_0", "taxonomy", "targets", 1, 0, 1, 100],
            ["coassembly_0", "nontarget_bin_sequences", "recovery", 1, 2, 3, 33.33],
            ["coassembly_0", "nontarget_unbin_sequences", "recovery", 0, 3, 3, 0],
            ["coassembly_0", "novel_sequences", "recovery", 1, 2, 3, 33.33],
            ["coassembly_0", "bins", "recovery", 1, 1, 2, 50],
        ], schema=SUMMARY_COLUMNS)

        observed_matches, observed_unmatched, observed_summary = evaluate(targets, binned, clusters, edges, recovered, recovered_bins)
        self.assertDataFrameEqual(expected_matches, observed_matches)
        self.assertDataFrameEqual(expected_unmatched, observed_unmatched)
        self.assertDataFrameEqual(expected_summary, observed_summary)

    def test_evaluate_script_multiple_coassemblies(self):
        targets = pl.DataFrame([
            ["S3.1", "sample_1.1", "AAA", 5, 10.0, "Root; old", 10],
            ["S3.1", "sample_2.1", "AAA", 5, 10.0, "Root; old", 10],
            ["S3.1", "sample_1.1", "CCC", 5, 10.0, "Root; old", 11],
            ["S3.1", "sample_3.1", "CCC", 5, 10.0, "Root; old", 11],
            ["S3.1", "sample_1.1", "EEE", 1, 2.0, "Root; old", 12],
            ["S3.1", "sample_3.1", "EEE", 1, 2.0, "Root; old", 12],
        ], schema=TARGET_COLUMNS)
        binned = pl.DataFrame([
            ["S3.1", "sample_1.1", "BBB", 5, 10.0, "Root; old", "oldgenome_1"],
            ["S3.1", "sample_3.1", "DDD", 5, 10.0, "Root; old", "oldgenome_2"],
        ], schema=APPRAISE_COLUMNS)
        clusters = pl.DataFrame([
            ["sample_1,sample_2", 2, 1, 100, "sample_1,sample_2,sample_3", "coassembly_0"],
            ["sample_1,sample_3", 2, 1, 100, "sample_1,sample_2,sample_3", "coassembly_1"],
        ], schema=CLUSTER_COLUMNS)
        edges = pl.DataFrame([
            ["match", 2, "sample_1.1,sample_2.1", "10"],
            ["match", 2, "sample_1.1,sample_3.1", "11"],
        ], schema=EDGE_COLUMNS)
        recovered = pl.DataFrame([
            ["S3.1", "coassembly_0-genome_1_transcripts", "AAA", 1, 2.0, "Root"],
            ["S3.1", "coassembly_0-genome_1_transcripts", "AAB", 1, 2.0, "Root"],
            ["S3.1", "coassembly_0-genome_1_transcripts", "BBB", 1, 2.0, "Root"],
            ["S3.1", "coassembly_1-genome_1_transcripts", "CCC", 1, 2.0, "Root"],
            ["S3.1", "coassembly_1-genome_1_transcripts", "CCD", 1, 2.0, "Root"],
            ["S3.1", "coassembly_1-genome_1_transcripts", "DDD", 1, 2.0, "Root"],
            ["S3.1", "coassembly_1-genome_1_transcripts", "EEE", 1, 2.0, "Root"],
        ], schema=SINGLEM_COLUMNS)
        recovered_bins = {
            "coassembly_0-genome_0": "genome_0.fna",
            "coassembly_0-genome_1": "genome_1.fna",
            "coassembly_1-genome_0": "genome_0.fna",
            "coassembly_1-genome_1": "genome_1.fna",
            }

        expected_matches = pl.DataFrame([
            ["coassembly_0", "S3.1", "AAA", "genome_1_transcripts", "10", None, "sample_1,sample_2", "Root"],
            ["coassembly_0", "S3.1", "BBB", "genome_1_transcripts", None, "oldgenome_1", "sample_1", "Root"],
            ["coassembly_1", "S3.1", "CCC", "genome_1_transcripts", "11", None, "sample_1,sample_3", "Root"],
            ["coassembly_1", "S3.1", "DDD", "genome_1_transcripts", None, "oldgenome_2", "sample_3", "Root"],
            ["coassembly_1", "S3.1", "EEE", "genome_1_transcripts", None, None, "sample_1,sample_3", "Root"],
        ], schema=OUTPUT_COLUMNS)
        expected_unmatched = pl.DataFrame([
            ["coassembly_0", "S3.1", "AAB", "genome_1_transcripts", None, None, None, "Root"],
            ["coassembly_1", "S3.1", "CCD", "genome_1_transcripts", None, None, None, "Root"],
        ], schema=OUTPUT_COLUMNS)
        expected_summary = pl.DataFrame([
            ["coassembly_0", "sequences", "targets", 1, 0, 1, 100],
            ["coassembly_0", "taxonomy", "targets", 1, 0, 1, 100],
            ["coassembly_0", "nontarget_bin_sequences", "recovery", 1, 2, 3, 33.33],
            ["coassembly_0", "nontarget_unbin_sequences", "recovery", 0, 3, 3, 0],
            ["coassembly_0", "novel_sequences", "recovery", 1, 2, 3, 33.33],
            ["coassembly_0", "bins", "recovery", 1, 1, 2, 50],
            ["coassembly_1", "sequences", "targets", 1, 0, 1, 100],
            ["coassembly_1", "taxonomy", "targets", 1, 0, 1, 100],
            ["coassembly_1", "nontarget_bin_sequences", "recovery", 1, 3, 4, 25.0],
            ["coassembly_1", "nontarget_unbin_sequences", "recovery", 1, 3, 4, 25.0],
            ["coassembly_1", "novel_sequences", "recovery", 1, 3, 4, 25.0],
            ["coassembly_1", "bins", "recovery", 1, 1, 2, 50],
        ], schema=SUMMARY_COLUMNS)

        observed_matches, observed_unmatched, observed_summary = evaluate(targets, binned, clusters, edges, recovered, recovered_bins)
        self.assertDataFrameEqual(expected_matches, observed_matches)
        self.assertDataFrameEqual(expected_unmatched, observed_unmatched)
        self.assertDataFrameEqual(expected_summary, observed_summary)

    def test_evaluate_script_subset_coassembly(self):
        targets = pl.DataFrame([
            ["S3.1", "sample_1.1", "AAA", 5, 10.0, "Root; old", 10],
            ["S3.1", "sample_2.1", "AAA", 5, 10.0, "Root; old", 10],
            ["S3.1", "sample_1.1", "CCC", 5, 10.0, "Root; old", 11],
            ["S3.1", "sample_3.1", "CCC", 5, 10.0, "Root; old", 11],
        ], schema=TARGET_COLUMNS)
        binned = pl.DataFrame([
            ["S3.1", "sample_1.1", "BBB", 5, 10.0, "Root; old", "oldgenome_1"],
            ["S3.1", "sample_3.1", "DDD", 5, 10.0, "Root; old", "oldgenome_2"],
        ], schema=APPRAISE_COLUMNS)
        clusters = pl.DataFrame([
            ["sample_1,sample_2,sample_3", 3, 1, 100, "sample_1,sample_2,sample_3", "coassembly_0"],
            ["sample_1,sample_2", 2, 1, 100, "sample_1,sample_2,sample_3", "coassembly_1"],
        ], schema=CLUSTER_COLUMNS)
        edges = pl.DataFrame([
            ["match", 2, "sample_1.1,sample_2.1", "10"],
            ["match", 2, "sample_1.1,sample_3.1", "11"],
        ], schema=EDGE_COLUMNS)
        recovered = pl.DataFrame([
            ["S3.1", "coassembly_0-genome_1_transcripts", "AAA", 1, 2.0, "Root"],
            ["S3.1", "coassembly_0-genome_1_transcripts", "AAB", 1, 2.0, "Root"],
            ["S3.1", "coassembly_0-genome_1_transcripts", "BBB", 1, 2.0, "Root"],
            ["S3.1", "coassembly_0-genome_1_transcripts", "CCC", 1, 2.0, "Root"],
            ["S3.1", "coassembly_0-genome_1_transcripts", "CCD", 1, 2.0, "Root"],
            ["S3.1", "coassembly_0-genome_1_transcripts", "DDD", 1, 2.0, "Root"],
            ["S3.1", "coassembly_1-genome_1_transcripts", "AAA", 1, 2.0, "Root"],
            ["S3.1", "coassembly_1-genome_1_transcripts", "AAB", 1, 2.0, "Root"],
            ["S3.1", "coassembly_1-genome_1_transcripts", "BBB", 1, 2.0, "Root"],
        ], schema=SINGLEM_COLUMNS)
        recovered_bins = {
            "coassembly_0-genome_0": "genome_0.fna",
            "coassembly_0-genome_1": "genome_1.fna",
            "coassembly_1-genome_0": "genome_0.fna",
            "coassembly_1-genome_1": "genome_1.fna",
            }

        expected_matches = pl.DataFrame([
            ["coassembly_0", "S3.1", "AAA", "genome_1_transcripts", "10", None, "sample_1,sample_2", "Root"],
            ["coassembly_0", "S3.1", "BBB", "genome_1_transcripts", None, "oldgenome_1", "sample_1", "Root"],
            ["coassembly_0", "S3.1", "CCC", "genome_1_transcripts", "11", None, "sample_1,sample_3", "Root"],
            ["coassembly_0", "S3.1", "DDD", "genome_1_transcripts", None, "oldgenome_2", "sample_3", "Root"],
            ["coassembly_1", "S3.1", "AAA", "genome_1_transcripts", "10", None, "sample_1,sample_2", "Root"],
            ["coassembly_1", "S3.1", "BBB", "genome_1_transcripts", None, "oldgenome_1", "sample_1", "Root"],
        ], schema=OUTPUT_COLUMNS)
        expected_unmatched = pl.DataFrame([
            ["coassembly_0", "S3.1", "AAB", "genome_1_transcripts", None, None, None, "Root"],
            ["coassembly_0", "S3.1", "CCD", "genome_1_transcripts", None, None, None, "Root"],
            ["coassembly_1", "S3.1", "AAB", "genome_1_transcripts", None, None, None, "Root"],
        ], schema=OUTPUT_COLUMNS)
        expected_summary = pl.DataFrame([
            ["coassembly_0", "sequences", "targets", 2, 0, 2, 100],
            ["coassembly_0", "taxonomy", "targets", 1, 0, 1, 100],
            ["coassembly_0", "nontarget_bin_sequences", "recovery", 2, 4, 6, 33.33],
            ["coassembly_0", "nontarget_unbin_sequences", "recovery", 0, 6, 6, 0],
            ["coassembly_0", "novel_sequences", "recovery", 2, 4, 6, 33.33],
            ["coassembly_0", "bins", "recovery", 1, 1, 2, 50],
            ["coassembly_1", "sequences", "targets", 1, 0, 1, 100],
            ["coassembly_1", "taxonomy", "targets", 1, 0, 1, 100],
            ["coassembly_1", "nontarget_bin_sequences", "recovery", 1, 2, 3, 33.33],
            ["coassembly_1", "nontarget_unbin_sequences", "recovery", 0, 3, 3, 0],
            ["coassembly_1", "novel_sequences", "recovery", 1, 2, 3, 33.33],
            ["coassembly_1", "bins", "recovery", 1, 1, 2, 50],
        ], schema=SUMMARY_COLUMNS)

        observed_matches, observed_unmatched, observed_summary = evaluate(targets, binned, clusters, edges, recovered, recovered_bins)
        self.assertDataFrameEqual(expected_matches, observed_matches)
        self.assertDataFrameEqual(expected_unmatched, observed_unmatched)
        self.assertDataFrameEqual(expected_summary, observed_summary)

    def test_evaluate_script_duplicate_sequences(self):
        targets = pl.DataFrame([
            ["S3.1", "sample_1.1", "AAA", 5, 10.0, "Root; old", 10],
            ["S3.1", "sample_3.1", "AAA", 5, 10.0, "Root; old", 10],
            ["S3.1", "sample_1.1", "CCC", 5, 10.0, "Root; old", 11],
            ["S3.1", "sample_2.1", "CCC", 5, 10.0, "Root; old", 11],
            ["S3.1", "sample_1.1", "YYY", 5, 10.0, "Root; old", 98],
            ["S3.1", "sample_2.1", "YYY", 5, 10.0, "Root; old", 98],
            ["S3.1", "sample_2.1", "ZZZ", 5, 10.0, "Root; old", 99],
            ["S3.1", "sample_3.1", "ZZZ", 5, 10.0, "Root; old", 99],
        ], schema=TARGET_COLUMNS)
        binned = pl.DataFrame([
            ["S3.1", "sample_3.1", "BBB", 5, 10.0, "Root; old", "oldgenome_1"],
            ["S3.1", "sample_1.1", "DDD", 5, 10.0, "Root; old", "oldgenome_2"],
        ], schema=APPRAISE_COLUMNS)
        clusters = pl.DataFrame([
            ["sample_1,sample_2,sample_3", 3, 2, 3000, "sample_1,sample_2,sample_3", "coassembly_0"],
            ["sample_1,sample_2", 2, 2, 2000, "sample_1,sample_2,sample_3", "coassembly_1"],
        ], schema=CLUSTER_COLUMNS)
        edges = pl.DataFrame([
            ["match", 2, "sample_1.1,sample_3.1", "10"],
            ["match", 2, "sample_1.1,sample_2.1", "11,98"],
            ["match", 2, "sample_2.1,sample_3.1", "99"],
        ], schema=EDGE_COLUMNS)
        recovered = pl.DataFrame([
            ["S3.1", "coassembly_0-genome_0_transcripts", "AAA", 1, 2.0, "Root"],
            ["S3.1", "coassembly_0-genome_1_transcripts", "AAA", 1, 2.0, "Root"],
            ["S3.1", "coassembly_0-genome_1_transcripts", "AAB", 1, 2.0, "Root"],
            ["S3.1", "coassembly_0-genome_1_transcripts", "BBB", 1, 2.0, "Root"],
            ["S3.1", "coassembly_1-genome_1_transcripts", "CCC", 1, 2.0, "Root"],
            ["S3.1", "coassembly_1-genome_1_transcripts", "CCD", 1, 2.0, "Root"],
            ["S3.1", "coassembly_1-genome_1_transcripts", "DDD", 1, 2.0, "Root"],
        ], schema=SINGLEM_COLUMNS)
        recovered_bins = {
            "coassembly_0-genome_0": "genome_0.fna",
            "coassembly_0-genome_1": "genome_1.fna",
            "coassembly_1-genome_0": "genome_0.fna",
            "coassembly_1-genome_1": "genome_1.fna",
            "coassembly_1-genome_2": "genome_2.fna",
            }

        expected_matches = pl.DataFrame([
            ["coassembly_0", "S3.1", "AAA", "genome_0_transcripts", "10", None, "sample_1,sample_3", "Root"],
            ["coassembly_0", "S3.1", "AAA", "genome_1_transcripts", "10", None, "sample_1,sample_3", "Root"],
            ["coassembly_0", "S3.1", "BBB", "genome_1_transcripts", None, "oldgenome_1", "sample_3", "Root"],
            ["coassembly_0", "S3.1", "CCC", None, "11", None, "sample_1,sample_2", "Root; old"],
            ["coassembly_0", "S3.1", "YYY", None, "98", None, "sample_1,sample_2", "Root; old"],
            ["coassembly_0", "S3.1", "ZZZ", None, "99", None, "sample_2,sample_3", "Root; old"],
            ["coassembly_1", "S3.1", "CCC", "genome_1_transcripts", "11", None, "sample_1,sample_2", "Root"],
            ["coassembly_1", "S3.1", "DDD", "genome_1_transcripts", None, "oldgenome_2", "sample_1", "Root"],
            ["coassembly_1", "S3.1", "YYY", None, "98", None, "sample_1,sample_2", "Root; old"],
        ], schema=OUTPUT_COLUMNS)
        expected_unmatched = pl.DataFrame([
            ["coassembly_0", "S3.1", "AAB", "genome_1_transcripts", None, None, None, "Root"],
            ["coassembly_1", "S3.1", "CCD", "genome_1_transcripts", None, None, None, "Root"],
        ], schema=OUTPUT_COLUMNS)
        expected_summary = pl.DataFrame([
            ["coassembly_0", "sequences", "targets", 1, 3, 4, 25],
            ["coassembly_0", "taxonomy", "targets", 1, 1, 2, 50],
            ["coassembly_0", "nontarget_bin_sequences", "recovery", 1, 3, 4, 25],
            ["coassembly_0", "nontarget_unbin_sequences", "recovery", 0, 4, 4, 0],
            ["coassembly_0", "novel_sequences", "recovery", 1, 3, 4, 25],
            ["coassembly_0", "bins", "recovery", 2, 0, 2, 100],
            ["coassembly_1", "sequences", "targets", 1, 1, 2, 50],
            ["coassembly_1", "taxonomy", "targets", 1, 1, 2, 50],
            ["coassembly_1", "nontarget_bin_sequences", "recovery", 1, 2, 3, 33.33],
            ["coassembly_1", "nontarget_unbin_sequences", "recovery", 0, 3, 3, 0],
            ["coassembly_1", "novel_sequences", "recovery", 1, 2, 3, 33.33],
            ["coassembly_1", "bins", "recovery", 1, 2, 3, 33.33],
        ], schema=SUMMARY_COLUMNS)

        observed_matches, observed_unmatched, observed_summary = evaluate(targets, binned, clusters, edges, recovered, recovered_bins)
        self.assertDataFrameEqual(expected_matches, observed_matches)
        self.assertDataFrameEqual(expected_unmatched, observed_unmatched)
        self.assertDataFrameEqual(expected_summary, observed_summary)

    def test_evaluate_script_other_sample_target(self):
        targets = pl.DataFrame([
            ["S3.1", "sample_1.1", "AAA", 5, 10.0, "Root; old", 10],
            ["S3.1", "sample_2.1", "AAA", 5, 10.0, "Root; old", 10],
            ["S3.1", "sample_1.1", "CCC", 1, 2.0, "Root; old", 11],
            ["S3.1", "sample_3.1", "CCC", 1, 2.0, "Root; old", 11],
        ], schema=TARGET_COLUMNS)
        binned = pl.DataFrame([
            ["S3.1", "sample_1.1", "BBB", 5, 10.0, "Root; old", "oldgenome_1"],
        ], schema=APPRAISE_COLUMNS)
        clusters = pl.DataFrame([
            ["sample_1,sample_2", 2, 1, 100, "sample_1,sample_2,sample_3", "coassembly_0"],
        ], schema=CLUSTER_COLUMNS)
        edges = pl.DataFrame([
            ["match", 2, "sample_1.1,sample_2.1", "10"],
        ], schema=EDGE_COLUMNS)
        recovered = pl.DataFrame([
            ["S3.1", "coassembly_0-genome_1_transcripts", "AAA", 1, 2.0, "Root"],
            ["S3.1", "coassembly_0-genome_1_transcripts", "CCC", 1, 2.0, "Root"],
        ], schema=SINGLEM_COLUMNS)
        recovered_bins = {
            "coassembly_0-genome_0": "genome_0.fna",
            "coassembly_0-genome_1": "genome_1.fna",
            }

        expected_matches = pl.DataFrame([
            ["coassembly_0", "S3.1", "AAA", "genome_1_transcripts", "10", None, "sample_1,sample_2", "Root"],
            ["coassembly_0", "S3.1", "CCC", "genome_1_transcripts", None, None, "sample_1", "Root"],
        ], schema=OUTPUT_COLUMNS)
        expected_unmatched = pl.DataFrame([
        ], schema=OUTPUT_COLUMNS)
        expected_summary = pl.DataFrame([
            ["coassembly_0", "sequences", "targets", 1, 0, 1, 100],
            ["coassembly_0", "taxonomy", "targets", 1, 0, 1, 100],
            ["coassembly_0", "nontarget_bin_sequences", "recovery", 0, 2, 2, 0],
            ["coassembly_0", "nontarget_unbin_sequences", "recovery", 1, 1, 2, 50.0],
            ["coassembly_0", "novel_sequences", "recovery", 0, 2, 2, 0],
            ["coassembly_0", "bins", "recovery", 1, 1, 2, 50],
        ], schema=SUMMARY_COLUMNS)

        observed_matches, observed_unmatched, observed_summary = evaluate(targets, binned, clusters, edges, recovered, recovered_bins)
        self.assertDataFrameEqual(expected_matches, observed_matches)
        self.assertDataFrameEqual(expected_unmatched, observed_unmatched)
        self.assertDataFrameEqual(expected_summary, observed_summary)


if __name__ == '__main__':
    unittest.main()



#####################
### no_genomes.py ###
#####################
# Author: Samuel Aroney

import polars as pl
import os

def processing(reads):
    print(f"Polars using {str(pl.threadpool_size())} threads")

    unbinned = (
        reads
        .with_columns(found_in = pl.lit(""))
        .filter(pl.col("gene") != "S3.18.EIF_2_alpha")
    )
    binned = unbinned.filter(False)

    return(binned, unbinned)

if __name__ == "__main__":
    os.environ["POLARS_MAX_THREADS"] = str(snakemake.threads)
    import polars as pl

    input_reads = snakemake.input.reads
    output_binned = snakemake.output.binned
    output_unbinned = snakemake.output.unbinned

    reads = []
    for read in input_reads:
        reads.append(pl.read_csv(read, separator="\t"))

    binned, unbinned = processing(pl.concat(reads))

    binned.write_csv(output_binned, separator="\t")
    unbinned.write_csv(output_unbinned, separator="\t")



#!/usr/bin/env python3

import unittest
import os
import shutil
import gzip
import subprocess
from ruamel.yaml import YAML

path_to_data = os.path.join(os.path.dirname(os.path.realpath(__file__)),'data')
path_to_conda = os.path.join(path_to_data,'.conda')

SAMPLE_READS_FORWARD = " ".join([
    os.path.join(path_to_data, "sample_1.1.fq"),
    os.path.join(path_to_data, "sample_2.1.fq"),
    os.path.join(path_to_data, "sample_3.1.fq"),
])
SAMPLE_READS_REVERSE = " ".join([
    os.path.join(path_to_data, "sample_1.2.fq"),
    os.path.join(path_to_data, "sample_2.2.fq"),
    os.path.join(path_to_data, "sample_3.2.fq"),
])
GENOMES = " ".join([os.path.join(path_to_data, "GB_GCA_013286235.1.fna")])
TWO_GENOMES = " ".join([
    os.path.join(path_to_data, "GB_GCA_013286235.1.fna"),
    os.path.join(path_to_data, "GB_GCA_013286235.2.fna"),
    ])

MOCK_COASSEMBLE = os.path.join(path_to_data, "mock_coassemble")
APPRAISE_BINNED = os.path.join(MOCK_COASSEMBLE, "appraise", "binned.otu_table.tsv")
APPRAISE_UNBINNED = os.path.join(MOCK_COASSEMBLE, "appraise", "unbinned.otu_table.tsv")
ELUSIVE_CLUSTERS = os.path.join(MOCK_COASSEMBLE, "target", "elusive_clusters.tsv")
ELUSIVE_CLUSTERS_TWO = os.path.join(MOCK_COASSEMBLE, 'target', 'elusive_clusters_two.tsv')


class Tests(unittest.TestCase):
    def setup_output_dir(self, output_dir):
        try:
            shutil.rmtree(output_dir)
        except FileNotFoundError:
            pass
        os.makedirs(output_dir)

    def test_update_sra_download_real(self):
        output_dir = os.path.join("example", "test_update_sra_download_real")
        self.setup_output_dir(output_dir)

        cmd = (
            f"ibis update "
            f"--assemble-unmapped "
            f"--forward SRR8334323 SRR8334324 "
            f"--sra "
            f"--genomes {GENOMES} "
            f"--coassemble-unbinned {os.path.join(MOCK_COASSEMBLE, 'appraise', 'unbinned_sra.otu_table.tsv')} "
            f"--coassemble-binned {os.path.join(MOCK_COASSEMBLE, 'appraise', 'binned_sra.otu_table.tsv')} "
            f"--coassemble-targets {os.path.join(MOCK_COASSEMBLE, 'target', 'targets.tsv')} "
            f"--coassemble-elusive-edges {os.path.join(MOCK_COASSEMBLE, 'target', 'elusive_edges.tsv')} "
            f"--coassemble-elusive-clusters {os.path.join(MOCK_COASSEMBLE, 'target', 'elusive_clusters_sra.tsv')} "
            f"--coassemble-summary {os.path.join(MOCK_COASSEMBLE, 'summary.tsv')} "
            f"--output {output_dir} "
            f"--conda-prefix {path_to_conda} "
        )
        subprocess.run(cmd, shell=True, check=True)

        config_path = os.path.join(output_dir, "config.yaml")
        self.assertTrue(os.path.exists(config_path))

        sra_1_path = os.path.join(output_dir, "coassemble", "sra", "SRR8334323_1.fastq.gz")
        self.assertTrue(os.path.exists(sra_1_path))
        with gzip.open(sra_1_path) as f:
            file = f.readline().decode()
            self.assertTrue("@SRR8334323.1 HS2:487:H80UEADXX:1:1101:1148:1986/1" in file)
            self.assertTrue("@SRR8334323.2 HS2:487:H80UEADXX:1:1101:1148:1986/2" not in file)

        self.assertTrue(os.path.exists(os.path.join(output_dir, "coassemble", "sra", "SRR8334323_2.fastq.gz")))
        self.assertTrue(os.path.exists(os.path.join(output_dir, "coassemble", "sra", "SRR8334324_1.fastq.gz")))
        self.assertTrue(os.path.exists(os.path.join(output_dir, "coassemble", "sra", "SRR8334324_2.fastq.gz")))

        sra_1_path = os.path.join(output_dir, "coassemble", "sra_qc", "SRR8334323_1.fastq.gz")
        self.assertTrue(os.path.exists(sra_1_path))
        with gzip.open(sra_1_path) as f:
            file = f.readline().decode()
            self.assertTrue("@SRR8334323.1 HS2:487:H80UEADXX:1:1101:1148:1986/1" in file)
            self.assertTrue("@SRR8334323.11 HS2:487:H80UEADXX:1:1101:2569:1958/1" not in file)
            self.assertTrue("@SRR8334323.15 HS2:487:H80UEADXX:1:1101:3302:1937/1" not in file)

        self.assertTrue(os.path.exists(os.path.join(output_dir, "coassemble", "sra_qc", "SRR8334323_2.fastq.gz")))
        self.assertTrue(os.path.exists(os.path.join(output_dir, "coassemble", "sra_qc", "SRR8334324_1.fastq.gz")))
        self.assertTrue(os.path.exists(os.path.join(output_dir, "coassemble", "sra_qc", "SRR8334324_2.fastq.gz")))

    def test_update_aviary_run_real(self):
        output_dir = os.path.join("example", "test_update_aviary_run_real")
        self.setup_output_dir(output_dir)

        cmd = (
            f"ibis update "
            f"--assemble-unmapped "
            f"--forward SRR8334323 SRR8334324 "
            f"--sra "
            f"--run-aviary "
            f"--cores 32 "
            f"--aviary-gtdbtk-dir /work/microbiome/db/gtdb/gtdb_release207_v2 "
            f"--aviary-checkm2-dir /work/microbiome/db/CheckM2_database "
            f"--genomes {GENOMES} "
            f"--coassemble-unbinned {os.path.join(MOCK_COASSEMBLE, 'appraise', 'unbinned_sra.otu_table.tsv')} "
            f"--coassemble-binned {os.path.join(MOCK_COASSEMBLE, 'appraise', 'binned_sra.otu_table.tsv')} "
            f"--coassemble-targets {os.path.join(MOCK_COASSEMBLE, 'target', 'targets.tsv')} "
            f"--coassemble-elusive-edges {os.path.join(MOCK_COASSEMBLE, 'target', 'elusive_edges.tsv')} "
            f"--coassemble-elusive-clusters {os.path.join(MOCK_COASSEMBLE, 'target', 'elusive_clusters_sra.tsv')} "
            f"--coassemble-summary {os.path.join(MOCK_COASSEMBLE, 'summary.tsv')} "
            f"--output {output_dir} "
            f"--conda-prefix {path_to_conda} "
            f"--snakemake-args '--profile mqsub --retries 1' "
        )
        subprocess.run(cmd, shell=True, check=True)

        config_path = os.path.join(output_dir, "config.yaml")
        self.assertTrue(os.path.exists(config_path))

        self.assertTrue(os.path.exists(os.path.join(output_dir, "coassemble", "sra", "SRR8334323_1.fastq.gz")))
        self.assertTrue(os.path.exists(os.path.join(output_dir, "coassemble", "sra", "SRR8334323_2.fastq.gz")))
        self.assertTrue(os.path.exists(os.path.join(output_dir, "coassemble", "sra", "SRR8334324_1.fastq.gz")))
        self.assertTrue(os.path.exists(os.path.join(output_dir, "coassemble", "sra", "SRR8334324_2.fastq.gz")))

        self.assertTrue(os.path.exists(os.path.join(output_dir, "coassemble", "coassemble", "coassembly_0", "assemble", "assembly", "final_contigs.fasta")))

        self.assertTrue(os.path.exists(os.path.join(output_dir, "coassemble", "coassemble", "coassembly_0", "recover", "bins", "checkm_minimal.tsv")))

    def test_update_aviary_run_real_large_assembly(self):
        output_dir = os.path.join("example", "test_update_aviary_run_real_large_assembly")
        self.setup_output_dir(output_dir)

        cmd = (
            f"ibis update "
            f"--assemble-unmapped "
            f"--forward SRR5753868 SRR5753874 "
            f"--sra "
            f"--genomes {GENOMES} "
            f"--run-aviary "
            f"--cores 32 "
            f"--aviary-gtdbtk-dir /work/microbiome/db/gtdb/gtdb_release207_v2 "
            f"--aviary-checkm2-dir /work/microbiome/db/CheckM2_database "
            f"--coassemble-unbinned {os.path.join(MOCK_COASSEMBLE, 'appraise', 'unbinned_sra.otu_table.tsv')} "
            f"--coassemble-binned {os.path.join(MOCK_COASSEMBLE, 'appraise', 'binned_sra.otu_table.tsv')} "
            f"--coassemble-targets {os.path.join(MOCK_COASSEMBLE, 'target', 'targets.tsv')} "
            f"--coassemble-elusive-edges {os.path.join(MOCK_COASSEMBLE, 'target', 'elusive_edges.tsv')} "
            f"--coassemble-elusive-clusters {os.path.join(MOCK_COASSEMBLE, 'target', 'elusive_clusters_sra_large.tsv')} "
            f"--coassemble-summary {os.path.join(MOCK_COASSEMBLE, 'summary.tsv')} "
            f"--output {output_dir} "
            f"--conda-prefix {path_to_conda} "
            f"--snakemake-args '--profile mqsub --retries 1' "
        )
        subprocess.run(cmd, shell=True, check=True)


        config_path = os.path.join(output_dir, "config.yaml")
        self.assertTrue(os.path.exists(config_path))

        self.assertTrue(os.path.exists(os.path.join(output_dir, "coassemble", "sra", "SRR5753868_1.fastq.gz")))
        self.assertTrue(os.path.exists(os.path.join(output_dir, "coassemble", "sra", "SRR5753868_2.fastq.gz")))
        self.assertTrue(os.path.exists(os.path.join(output_dir, "coassemble", "sra", "SRR5753874_1.fastq.gz")))
        self.assertTrue(os.path.exists(os.path.join(output_dir, "coassemble", "sra", "SRR5753874_2.fastq.gz")))

        self.assertTrue(os.path.exists(os.path.join(output_dir, "coassemble", "coassemble", "coassembly_0", "assemble", "assembly", "final_contigs.fna")))

        config_path = os.path.join(output_dir, "coassemble", "coassemble", "coassembly_0", "assemble", "config.yaml")
        self.assertTrue(os.path.exists(config_path))
        with open(config_path) as f:
            config = YAML().load(f)
        self.assertTrue(config["use_megahit"])

        self.assertTrue(os.path.exists(os.path.join(output_dir, "coassemble", "coassemble", "coassembly_0", "recover", "bins", "checkm_minimal.tsv")))

    def test_update_specific_coassembly_sra(self):
        output_dir = os.path.join("example", "test_update_specific_coassembly_sra")
        self.setup_output_dir(output_dir)

        cmd = (
            f"ibis update "
            f"--forward SRR8334323 SRR8334324 "
            f"--sra "
            f"--genomes {GENOMES} "
            f"--coassemble-unbinned {os.path.join(MOCK_COASSEMBLE, 'appraise', 'unbinned_sra.otu_table.tsv')} "
            f"--coassemble-binned {os.path.join(MOCK_COASSEMBLE, 'appraise', 'binned_sra.otu_table.tsv')} "
            f"--coassemble-targets {os.path.join(MOCK_COASSEMBLE, 'target', 'targets.tsv')} "
            f"--coassemble-elusive-edges {os.path.join(MOCK_COASSEMBLE, 'target', 'elusive_edges.tsv')} "
            f"--coassemble-elusive-clusters {os.path.join(MOCK_COASSEMBLE, 'target', 'elusive_clusters_sra_two.tsv')} "
            f"--coassemble-summary {os.path.join(MOCK_COASSEMBLE, 'summary.tsv')} "
            f"--coassemblies coassembly_0 "
            f"--output {output_dir} "
            f"--conda-prefix {path_to_conda} "
        )
        subprocess.run(cmd, shell=True, check=True)

        config_path = os.path.join(output_dir, "config.yaml")
        self.assertTrue(os.path.exists(config_path))

        self.assertTrue(os.path.exists(os.path.join(output_dir, "coassemble", "sra", "SRR8334323_1.fastq.gz")))
        self.assertTrue(os.path.exists(os.path.join(output_dir, "coassemble", "sra", "SRR8334323_2.fastq.gz")))
        self.assertTrue(os.path.exists(os.path.join(output_dir, "coassemble", "sra", "SRR8334324_1.fastq.gz")))
        self.assertTrue(os.path.exists(os.path.join(output_dir, "coassemble", "sra", "SRR8334324_2.fastq.gz")))

        self.assertFalse(os.path.exists(os.path.join(output_dir, "coassemble", "sra", "SRR8334320_1.fastq.gz")))
        self.assertFalse(os.path.exists(os.path.join(output_dir, "coassemble", "sra", "SRR8334320_2.fastq.gz")))
        self.assertFalse(os.path.exists(os.path.join(output_dir, "coassemble", "sra", "SRR8334321_1.fastq.gz")))
        self.assertFalse(os.path.exists(os.path.join(output_dir, "coassemble", "sra", "SRR8334321_2.fastq.gz")))


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3

import unittest
import os
os.environ["POLARS_MAX_THREADS"] = "1"
import polars as pl
from polars.testing import assert_frame_equal
from ibis.workflow.scripts.aviary_commands import pipeline

ELUSIVE_CLUSTERS_COLUMNS={
    "samples": str,
    "length": int,
    "total_targets": int,
    "total_size": int,
    "recover_samples": str,
    "coassembly": str,
    }
COMMANDS_COLUMNS={
    "assemble": str,
    "recover": str,
}

reads_1 = {
    "1": "1_1.fq.gz",
    "2": "2_1.fq.gz",
    "3": "3_1.fq.gz",
    "4": "4_1.fq.gz",
    "5": "5_1.fq.gz",
    "6": "6_1.fq.gz",
}
reads_2 = {
    "1": "1_2.fq.gz",
    "2": "2_2.fq.gz",
    "3": "3_2.fq.gz",
    "4": "4_2.fq.gz",
    "5": "5_2.fq.gz",
    "6": "6_2.fq.gz",
}

class Tests(unittest.TestCase):
    def assertDataFrameEqual(self, a, b):
        assert_frame_equal(a, b, check_dtype=False)

    def test_aviary_commands(self):
        elusive_clusters = pl.DataFrame([
            ["1,2,3", 3, 2, 2000, "1,2,3,4,5,6", "coassembly_0"],
            ["4,5,6", 3, 1, 3000, "4,5,6", "coassembly_2"],
        ], schema=ELUSIVE_CLUSTERS_COLUMNS)
        output_dir = "test_output_dir"
        threads = 10
        memory = 50


        expected_commands = pl.DataFrame([
            [
                f"aviary assemble "
                f"-1 1_1.fq.gz 2_1.fq.gz 3_1.fq.gz "
                f"-2 1_2.fq.gz 2_2.fq.gz 3_2.fq.gz "
                f"--output {output_dir}/coassemble/coassembly_0/assemble "
                f"-n {threads} -t {threads} -m {memory} "
                f"&> {output_dir}/coassemble/logs/coassembly_0_assemble.log ",

                f"aviary recover --assembly {output_dir}/coassemble/coassembly_0/assemble/assembly/final_contigs.fasta "
                f"-1 1_1.fq.gz 2_1.fq.gz 3_1.fq.gz 4_1.fq.gz 5_1.fq.gz 6_1.fq.gz "
                f"-2 1_2.fq.gz 2_2.fq.gz 3_2.fq.gz 4_2.fq.gz 5_2.fq.gz 6_2.fq.gz "
                f"--output {output_dir}/coassemble/coassembly_0/recover "
                f"-n {threads//2} -t {threads//2} -m {memory//2} "
                f"&> {output_dir}/coassemble/logs/coassembly_0_recover.log ",
            ],
            [
                f"aviary assemble "
                f"-1 4_1.fq.gz 5_1.fq.gz 6_1.fq.gz "
                f"-2 4_2.fq.gz 5_2.fq.gz 6_2.fq.gz "
                f"--output {output_dir}/coassemble/coassembly_2/assemble "
                f"-n {threads} -t {threads} -m {memory} "
                f"&> {output_dir}/coassemble/logs/coassembly_2_assemble.log ",

                f"aviary recover --assembly {output_dir}/coassemble/coassembly_2/assemble/assembly/final_contigs.fasta "
                f"-1 4_1.fq.gz 5_1.fq.gz 6_1.fq.gz "
                f"-2 4_2.fq.gz 5_2.fq.gz 6_2.fq.gz "
                f"--output {output_dir}/coassemble/coassembly_2/recover "
                f"-n {threads//2} -t {threads//2} -m {memory//2} "
                f"&> {output_dir}/coassemble/logs/coassembly_2_recover.log ",
            ],
        ], schema=COMMANDS_COLUMNS, orient="row")

        observed_commands = pipeline(elusive_clusters, reads_1, reads_2, output_dir, threads, memory)
        self.assertDataFrameEqual(expected_commands, observed_commands)

    def test_aviary_commands_fast(self):
        elusive_clusters = pl.DataFrame([
            ["1,2,3", 3, 2, 2000, "1,2,3,4,5,6", "coassembly_0"],
            ["4,5,6", 3, 1, 3000, "4,5,6", "coassembly_2"],
        ], schema=ELUSIVE_CLUSTERS_COLUMNS)
        output_dir = "test_output_dir"
        threads = 10
        memory = 50


        expected_commands = pl.DataFrame([
            [
                f"aviary assemble "
                f"-1 1_1.fq.gz 2_1.fq.gz 3_1.fq.gz "
                f"-2 1_2.fq.gz 2_2.fq.gz 3_2.fq.gz "
                f"--output {output_dir}/coassemble/coassembly_0/assemble "
                f"-n {threads} -t {threads} -m {memory} "
                f"&> {output_dir}/coassemble/logs/coassembly_0_assemble.log ",

                f"aviary recover --assembly {output_dir}/coassemble/coassembly_0/assemble/assembly/final_contigs.fasta "
                f"-1 1_1.fq.gz 2_1.fq.gz 3_1.fq.gz 4_1.fq.gz 5_1.fq.gz 6_1.fq.gz "
                f"-2 1_2.fq.gz 2_2.fq.gz 3_2.fq.gz 4_2.fq.gz 5_2.fq.gz 6_2.fq.gz "
                f"--output {output_dir}/coassemble/coassembly_0/recover "
                f"--workflow recover_mags_no_singlem "
                f"--skip-binners maxbin concoct rosella "
                f"--skip-abundances "
                f"--refinery-max-iterations 0 "
                f"-n {threads//2} -t {threads//2} -m {memory//2} "
                f"&> {output_dir}/coassemble/logs/coassembly_0_recover.log ",
            ],
            [
                f"aviary assemble "
                f"-1 4_1.fq.gz 5_1.fq.gz 6_1.fq.gz "
                f"-2 4_2.fq.gz 5_2.fq.gz 6_2.fq.gz "
                f"--output {output_dir}/coassemble/coassembly_2/assemble "
                f"-n {threads} -t {threads} -m {memory} "
                f"&> {output_dir}/coassemble/logs/coassembly_2_assemble.log ",

                f"aviary recover --assembly {output_dir}/coassemble/coassembly_2/assemble/assembly/final_contigs.fasta "
                f"-1 4_1.fq.gz 5_1.fq.gz 6_1.fq.gz "
                f"-2 4_2.fq.gz 5_2.fq.gz 6_2.fq.gz "
                f"--output {output_dir}/coassemble/coassembly_2/recover "
                f"--workflow recover_mags_no_singlem "
                f"--skip-binners maxbin concoct rosella "
                f"--skip-abundances "
                f"--refinery-max-iterations 0 "
                f"-n {threads//2} -t {threads//2} -m {memory//2} "
                f"&> {output_dir}/coassemble/logs/coassembly_2_recover.log ",
            ],
        ], schema=COMMANDS_COLUMNS, orient="row")

        observed_commands = pipeline(elusive_clusters, reads_1, reads_2, output_dir, threads, memory, fast=True)
        self.assertDataFrameEqual(expected_commands, observed_commands)


if __name__ == '__main__':
    unittest.main()



########################
### cluster_graph.py ###
########################
# Author: Samuel Aroney

import polars as pl
import itertools
import os
import logging

OUTPUT_COLUMNS={
    "samples": str,
    "length": int,
    "total_targets": int,
    "total_size": int,
    "recover_samples": str,
    "coassembly": str,
    }

def join_list_subsets(df1, df2):
    """
    Join two DataFrames/LazyFrames by strict subset of list column
    """
    df2 = (
        df2.select(
            right_samples = pl.col("samples"),
            extra_targets = pl.col("target_ids"),
            length = pl.col("cluster_size").cast(pl.UInt32),
            )
    )

    output = (
        df1
        .select(
            pl.col("samples"),
            pl.col("samples_hash"),
            pl.col("length"),
            )
        .join(
            df2,
            on=["length"],
            how="left",
        )
        .filter(pl.col("samples").list.set_intersection("right_samples").list.lengths() >= pl.col("length"))
        .groupby("samples_hash")
        .agg(pl.col("extra_targets").flatten())
        .select(
            pl.col("samples_hash"),
            pl.col("extra_targets").list.unique().fill_null([]),
            )
    )

    return (
        df1
        .join(output, on="samples_hash", how="left")
        .with_columns(pl.col("extra_targets").fill_null([]))
        )

def accumulate_clusters(x):
    clustered_samples = {}
    choices = []
    for cluster in x:
        cluster_size = len(cluster)
        if cluster_size not in clustered_samples:
            clustered_samples[cluster_size] = set()
        if all([s not in clustered_samples[cluster_size] for s in cluster]):
            choices.append(True)
            clustered_samples[cluster_size].update(cluster)
        else:
            choices.append(False)

    return pl.Series(choices, dtype=pl.Boolean)

def find_recover_candidates(df, samples_df, MAX_RECOVERY_SAMPLES=20):
    samples_df = samples_df.explode("target_ids")

    output = (
        df
        .select("samples_hash", "target_ids")
        .explode("target_ids")
        .join(samples_df, on="target_ids", how="left")
        .groupby("samples_hash", "recover_candidates")
        .agg(pl.count())
        .sort("count", descending=True)
        .groupby("samples_hash")
        .head(MAX_RECOVERY_SAMPLES)
        .groupby("samples_hash")
        .agg("recover_candidates")
    )

    return df.join(output, on="samples_hash", how="left")

def pipeline(
        elusive_edges,
        read_size,
        MAX_COASSEMBLY_SIZE=None,
        MAX_COASSEMBLY_SAMPLES=2,
        MIN_COASSEMBLY_SAMPLES=2,
        MAX_RECOVERY_SAMPLES=20,
        MIN_CLUSTER_TARGETS=1,
        MAX_SAMPLES_COMBINATIONS=100,
        EXCLUDE_COASSEMBLIES=[]):

    logging.info(f"Polars using {str(pl.threadpool_size())} threads")

    if len(elusive_edges) == 0:
        logging.warning("No elusive edges found")
        return pl.DataFrame(schema=OUTPUT_COLUMNS)

    is_pooled = any(elusive_edges["style"] == "pool")

    with pl.StringCache():
        elusive_edges = (
            elusive_edges
            .lazy()
            .with_columns(
                pl.col("samples")
                    .str.split(",")
                    .list.eval(
                        pl.when(pl.all_horizontal(pl.element().is_in(read_size.get_column("sample"))))
                        .then(pl.element())
                        .otherwise(pl.element().str.replace(r"(_|\.)1$", ""))
                        )
                    .cast(pl.List(pl.Categorical)),
                pl.col("target_ids")
                    .str.split(",")
                    .cast(pl.List(pl.UInt32)),
                )
            .with_columns(
                samples_hash = pl.col("samples").list.sort().hash(),
                length = pl.col("samples").list.lengths()
                )
        )

        read_size = (
            read_size
            .lazy()
            .select(
                "read_size",
                samples = pl.col("sample").cast(pl.Categorical),
                )
        )

        if EXCLUDE_COASSEMBLIES:
            excluded_coassemblies = (
                pl.LazyFrame({"samples": EXCLUDE_COASSEMBLIES})
                .with_columns(
                    samples_hash = pl.col("samples")
                        .str.split(",")
                        .cast(pl.List(pl.Categorical))
                        .list.sort().hash()
                    )
            )
        else:
            excluded_coassemblies = pl.LazyFrame(schema={"samples_hash": pl.UInt64})

        if MAX_COASSEMBLY_SAMPLES == 1:
            logging.info("Skipping clustering, using single-sample clusters")
            clusters = [
                elusive_edges
                .explode("samples")
                .groupby("samples")
                .agg(pl.col("target_ids").flatten())
                .with_columns(
                    pl.concat_list(pl.col("samples")),
                    pl.col("target_ids").list.sort().list.unique(),
                    samples_hash = pl.concat_list(pl.col("samples")).list.sort().hash(),
                )
            ]
        else:
            logging.info("Forming candidate sample clusters")
            clusters = [
                elusive_edges
                .filter(pl.col("style") == "match")
                .filter(pl.col("cluster_size") >= MIN_COASSEMBLY_SAMPLES)
                .filter(pl.col("target_ids").list.lengths() >= MIN_CLUSTER_TARGETS)
                .select("samples", "target_ids", samples_hash = pl.col("samples").list.sort().hash())
            ]

            if is_pooled:
                clusters.append(
                    elusive_edges
                    .filter(pl.col("style") == "pool")
                    .filter(pl.col("cluster_size") >= MIN_COASSEMBLY_SAMPLES)
                    # Prevent combinatorial explosion (also, large clusters are less useful for distinguishing between clusters)
                    .filter(pl.col("samples").list.lengths() < MAX_SAMPLES_COMBINATIONS)
                    .with_columns(
                        samples_combinations = pl.struct(["length", "cluster_size"]).apply(
                            lambda x: [i for i in itertools.combinations(range(x["length"]), x["cluster_size"])],
                            return_dtype=pl.List(pl.List(pl.Int64)),
                            )
                        )
                    .explode("samples_combinations")
                    .with_columns(pl.col("samples").list.take(pl.col("samples_combinations")))
                    .with_columns(samples_hash = pl.col("samples").list.sort().hash())
                    .select("samples_hash", "samples", "target_ids", "cluster_size")
                    .groupby("samples_hash")
                    .agg(
                        pl.first("samples"),
                        pl.col("target_ids").flatten(),
                        pl.first("cluster_size"),
                        )
                    .filter(pl.col("target_ids").list.lengths() >= MIN_CLUSTER_TARGETS)
                    .select("samples", "target_ids", "samples_hash")
                )

        sample_targets = (
            elusive_edges
            .select("target_ids", recover_candidates = pl.col("samples"))
            .explode("recover_candidates")
            .explode("target_ids")
            .unique()
            .groupby("recover_candidates")
            .agg("target_ids")
        )

        logging.info("Filtering clusters (each sample restricted to only once per cluster size)")
        clusters = (
            pl.concat(clusters)
            .join(excluded_coassemblies, on="samples_hash", how="anti")
            .with_columns(length = pl.col("samples").list.lengths())
            .explode("samples")
            .join(read_size, on="samples", how="left")
            .groupby("samples_hash")
            .agg(
                pl.col("samples"),
                pl.first("length"),
                pl.first("target_ids"),
                total_size = pl.sum("read_size"),
                )
            .filter(
                pl.when(MAX_COASSEMBLY_SIZE is not None)
                .then(pl.col("total_size") <= MAX_COASSEMBLY_SIZE)
                .otherwise(True)
                )
            .with_columns(
                total_targets = pl.col("target_ids").list.lengths(),
            )
            .sort("total_targets", "total_size", descending=[True, False])
            .with_columns(
                unique_samples = 
                    pl.col("samples")
                    .map(accumulate_clusters, return_dtype=pl.Boolean),
                )
            .filter(pl.col("unique_samples"))
            .drop("unique_samples")
            .collect(streaming=True)
            .pipe(
                join_list_subsets,
                df2=elusive_edges
                    .filter(pl.col("style") == "pool")
                    .filter(pl.col("samples").list.lengths() >= MAX_SAMPLES_COMBINATIONS)
                    .collect(streaming=True),
                )
            .with_columns(pl.concat_list("target_ids", "extra_targets").list.unique())
            .pipe(
                find_recover_candidates,
                sample_targets.collect(streaming=True),
                MAX_RECOVERY_SAMPLES=MAX_RECOVERY_SAMPLES,
                )
            .with_columns(
                pl.col("samples").cast(pl.List(pl.Utf8)).list.sort().list.join(","),
                recover_samples = pl.concat_list(pl.col("samples"), "recover_candidates")
                .list.unique(maintain_order=True)
                .list.head(MAX_RECOVERY_SAMPLES)
                .cast(pl.List(pl.Utf8))
                .list.sort()
                .list.join(","),
                total_targets = pl.col("target_ids").list.lengths(),
                )
            .sort("total_targets", "total_size", descending=[True, False])
            .with_row_count("coassembly")
            .select(
                "samples", "length", "total_targets", "total_size", "recover_samples",
                coassembly = pl.lit("coassembly_") + pl.col("coassembly").cast(pl.Utf8)
                )
        )

    logging.info("Done")

    return clusters

if __name__ == "__main__":
    os.environ["POLARS_MAX_THREADS"] = str(snakemake.threads)
    import polars as pl

    logging.basicConfig(
        level=logging.INFO,
        format='%(asctime)s %(levelname)s: %(message)s',
        datefmt='%Y/%m/%d %I:%M:%S %p'
        )

    MAX_COASSEMBLY_SIZE = snakemake.params.max_coassembly_size * 10**9 if snakemake.params.max_coassembly_size else None
    MAX_COASSEMBLY_SAMPLES = snakemake.params.max_coassembly_samples
    MIN_COASSEMBLY_SAMPLES = snakemake.params.num_coassembly_samples
    MAX_RECOVERY_SAMPLES = snakemake.params.max_recovery_samples
    EXCLUDE_COASSEMBLIES = snakemake.params.exclude_coassemblies
    elusive_edges_path = snakemake.input.elusive_edges
    read_size_path = snakemake.input.read_size
    elusive_clusters_path = snakemake.output.elusive_clusters

    elusive_edges = pl.read_csv(elusive_edges_path, separator="\t", dtypes={"target_ids": str})
    read_size = pl.read_csv(read_size_path, has_header=False, new_columns=["sample", "read_size"])

    if elusive_edges.height > 10**4:
        min_cluster_targets = 10
    else:
        min_cluster_targets = 1

    clusters = pipeline(
        elusive_edges,
        read_size,
        MAX_COASSEMBLY_SIZE=MAX_COASSEMBLY_SIZE,
        MAX_COASSEMBLY_SAMPLES=MAX_COASSEMBLY_SAMPLES,
        MIN_COASSEMBLY_SAMPLES=MIN_COASSEMBLY_SAMPLES,
        MAX_RECOVERY_SAMPLES=MAX_RECOVERY_SAMPLES,
        EXCLUDE_COASSEMBLIES=EXCLUDE_COASSEMBLIES,
        MIN_CLUSTER_TARGETS=min_cluster_targets,
        MAX_SAMPLES_COMBINATIONS=100,
        )
    clusters.write_csv(elusive_clusters_path, separator="\t")



#################################
### collect_reference_bins.py ###
#################################
# Author: Samuel Aroney

import polars as pl
import os
import numpy as np
import extern

def trimmed_mean(data, trim=0.1):
    cut = int(np.floor(len(data) * trim))
    if cut == 0:
        return np.mean(data)
    else:
        a = sorted(data)
        return np.mean(a[cut:-cut])

def pipeline(appraise_binned, appraise_unbinned, sample, MIN_APPRAISED=0.1, TRIM_FRACTION=0.1):
    print(f"Polars using {str(pl.threadpool_size())} threads")

    appraise_binned = appraise_binned.with_columns(
        pl.col("sample").cast(str),
        sample_remove_suffix1 = pl.col("sample").str.replace(r"\.1$", ""),
        sample_remove_suffix2 = pl.col("sample").str.replace(r"_1$", ""),
    ).filter(
        (pl.col("sample") == sample) | (pl.col("sample_remove_suffix1") == sample) | (pl.col("sample_remove_suffix2") == sample)
    )

    appraise_unbinned = appraise_unbinned.with_columns(
        pl.col("sample").cast(str),
        sample_remove_suffix1 = pl.col("sample").str.replace(r"\.1$", ""),
        sample_remove_suffix2 = pl.col("sample").str.replace(r"_1$", ""),
    ).filter(
        (pl.col("sample") == sample) | (pl.col("sample_remove_suffix1") == sample) | (pl.col("sample_remove_suffix2") == sample)
    )

    num_binned = sum(appraise_binned.get_column("num_hits").to_list())
    num_unbinned = sum(appraise_unbinned.get_column("num_hits").to_list())
    try:
        perc_binned = num_binned / (num_binned + num_unbinned)
    except ZeroDivisionError:
        return set()

    if perc_binned < MIN_APPRAISED:
        return set()

    reference_bins = appraise_binned.with_columns(
        pl.col("found_in").str.split(",")
    ).explode(
        "found_in"
    ).with_columns(
        pl.col("found_in").str.replace("_protein$", "")
    ).groupby(
        ["gene", "found_in"]
    ).agg(
        pl.col("coverage").sum()
    ).pivot(
        values="coverage", index="gene", columns="found_in", aggregate_function=None
    ).melt(
        id_vars="gene", variable_name="found_in", value_name="coverage"
    ).fill_null(0
    ).groupby(
        "found_in"
    ).agg(
        (pl.col("coverage").len() * TRIM_FRACTION).floor().cast(int).alias("cut"),
        pl.col("coverage")
    ).with_columns(
        pl.col("coverage").list.sort().list.slice(
            pl.col("cut"), pl.col("coverage").list.lengths() - 2 * pl.col("cut")
            ).list.mean()
    ).filter(
        pl.col("coverage") > 0
    ).get_column("found_in"
    ).to_list()

    return set(reference_bins)

if __name__ == "__main__":
    os.environ["POLARS_MAX_THREADS"] = str(snakemake.threads)
    import polars as pl

    binned_path = snakemake.input.appraise_binned
    unbinned_path = snakemake.input.appraise_unbinned
    genomes = snakemake.params.genomes
    sample = snakemake.params.sample
    MIN_APPRAISED = snakemake.params.min_appraised
    sample_read = snakemake.wildcards.read
    output_path = snakemake.output

    appraise_binned = pl.read_csv(binned_path, separator="\t")
    appraise_unbinned = pl.read_csv(unbinned_path, separator="\t")

    reference_bins = pipeline(appraise_binned, appraise_unbinned, sample, MIN_APPRAISED=MIN_APPRAISED)

    if len(reference_bins) == 0:
        print(f"Warning: No reference bins found for {sample_read}")
        cmd = f"touch {output_path}"
        extern.run(cmd)
    else:
        for bin in reference_bins:
            cmd = f"cat {genomes[bin]} >> {output_path}"
            extern.run(cmd)


#########################
### target_elusive.py ###
#########################
# Author: Samuel Aroney

import polars as pl
import os
import logging

EDGES_COLUMNS={
    "style": str,
    "cluster_size": int,
    "samples": str,
    "target_ids": str,
    }

def pipeline(
    unbinned,
    MIN_COASSEMBLY_COVERAGE=10,
    TAXA_OF_INTEREST="",
    MAX_COASSEMBLY_SAMPLES=2):

    logging.info(f"Polars using {str(pl.threadpool_size())} threads")

    if len(unbinned) == 0:
        logging.warning("No unbinned sequences found")
        return unbinned.rename({"found_in": "target"}), pl.DataFrame(schema=EDGES_COLUMNS)

    if MAX_COASSEMBLY_SAMPLES < 2:
        # Set to 2 to produce paired edges
        MAX_COASSEMBLY_SAMPLES = 2

    # Filter TAXA_OF_INTEREST
    if TAXA_OF_INTEREST:
        logging.info(f"Filtering for taxa of interest: {TAXA_OF_INTEREST}")
        unbinned = unbinned.filter(
            pl.col("taxonomy").str.contains(TAXA_OF_INTEREST, literal=True)
        )

    logging.info("Grouping hits by marker gene sequences to form targets")
    unbinned = (
        unbinned
        .drop("found_in")
        .with_row_count("target")
        .select(
            "gene", "sample", "sequence", "num_hits", "coverage", "taxonomy",
            pl.first("target").over(["gene", "sequence"]).rank("dense") - 1,
            )
        .with_columns(
            pl.col("target").cast(pl.Utf8)
            )
    )

    def process_groups(df):
        if df.height == 1:
            return pl.DataFrame(schema={"style": str, "cluster_size": pl.Int64, "samples": str, "target": str})

        # Direct matching samples in pairs with coverage > MIN_COASSEMBLY_COVERAGE
        dfs = [
            df
            .join(df, how="cross", suffix="_2")
            .filter(
                pl.col("samples").list.last().str.encode("hex") < pl.col("samples_2").list.first().str.encode("hex")
                )
            .select([
                "target",
                pl.col("coverage") + pl.col("coverage_2").alias("coverage"),
                pl.col("samples").list.concat(pl.col("samples_2")),
                ])
            .filter(pl.col("samples").list.lengths() > 1)
            .filter(pl.col("coverage") > MIN_COASSEMBLY_COVERAGE)
            .select(
                pl.lit("match").alias("style"),
                pl.lit(2).cast(pl.Int64).alias("cluster_size"),
                pl.col("samples").list.join(","),
                pl.col("target"),
                )
        ]

        # Pool samples with coverage > MIN_COASSEMBLY_COVERAGE / N for clusters of size N: 3 to MAX_COASSEMBLY_SAMPLES
        cluster_sizes = pl.DataFrame({"cluster_size": range(3, MAX_COASSEMBLY_SAMPLES+1)})
        dfs.append(
            df
            .join(cluster_sizes, how="cross")
            .filter(pl.col("coverage") > float(MIN_COASSEMBLY_COVERAGE) / pl.col("cluster_size").cast(float))
            .groupby("target", "cluster_size")
            .agg(pl.concat_list("samples").flatten())
            .filter(pl.col("samples").list.lengths() >= pl.col("cluster_size"))
            .select(
                pl.lit("pool").alias("style"),
                pl.col("cluster_size"),
                pl.col("samples").list.sort().list.join(","),
                pl.col("target"),
                )
        )

        return pl.concat(dfs)

    logging.info("Grouping targets into paired matches and pooled samples for clusters of size 3+")
    sparse_edges = (
        unbinned
        .select(
            "target",
            "coverage",
            samples = pl.col("sample").apply(lambda x: [x], return_dtype=pl.List(pl.Utf8)),
            )
        .groupby("target")
        .apply(process_groups)
        .groupby(["style", "cluster_size", "samples"])
        .agg(target_ids = pl.col("target").sort().str.concat(","))
    )

    return unbinned, sparse_edges

if __name__ == "__main__":
    os.environ["POLARS_MAX_THREADS"] = str(snakemake.threads)
    import polars as pl

    logging.basicConfig(
        level=logging.INFO,
        format='%(asctime)s %(levelname)s: %(message)s',
        datefmt='%Y/%m/%d %I:%M:%S %p'
        )

    MIN_COASSEMBLY_COVERAGE = snakemake.params.min_coassembly_coverage
    MAX_COASSEMBLY_SAMPLES = snakemake.params.max_coassembly_samples
    TAXA_OF_INTEREST = snakemake.params.taxa_of_interest
    unbinned_path = snakemake.input.unbinned
    targets_path = snakemake.output.output_targets
    edges_path = snakemake.output.output_edges

    unbinned = pl.read_csv(unbinned_path, separator="\t")

    targets, edges = pipeline(
        unbinned,
        MIN_COASSEMBLY_COVERAGE=MIN_COASSEMBLY_COVERAGE,
        TAXA_OF_INTEREST=TAXA_OF_INTEREST,
        MAX_COASSEMBLY_SAMPLES=MAX_COASSEMBLY_SAMPLES,
        )
    targets.write_csv(targets_path, separator="\t")
    edges.write_csv(edges_path, separator="\t")



###############################
### separate_SingleM_seq.py ###
###############################
# Author: Samuel Aroney

import pandas as pd
import re
from Bio.Seq import Seq
from Bio.SeqRecord import SeqRecord
from Bio import SeqIO

otu_table = pd.read_csv(snakemake.input.otu_table, sep="\t")

otu_table["id"] = otu_table.index
otu_table["id"] = otu_table["id"].astype(str)

for gene in otu_table["gene"].unique():
    gene_table = otu_table[otu_table["gene"] == gene]
    records = [SeqRecord(Seq(re.sub("-", "", gene_table.iloc[i]["sequence"])), id = gene_table.iloc[i]["id"]) for i in range(len(gene_table))]
    SeqIO.write(records, snakemake.params.output_dir + "/" + gene + ".fasta", 'fasta')

otu_table.to_csv(snakemake.output.id_otu_table, sep="\t", index=False)
